#include "PX_Firework.h"
#include "PainterEngine_Startup.h"

enum fw_state
{
	STATE_MOVING,
	STATE_TRIGGING,
	STATE_ENERGY_STORING,
	STATE_BLOOMING,
};

struct fw_points_t
{
	px_float x;
	px_float y;
	px_float speed_x;
	px_float speed_y;
	px_float size;
};

struct fw_sparks_t
{
	px_color color;
	struct fw_points_t points[16];
	px_float size;
	px_float speed_x;
	px_float speed_y;
};

struct fw_t
{
	enum fw_state state;
	struct fw_sparks_t sparks[256];
	px_int elapsed;
	px_point bloom_coord;
	px_point start_coord;
};

px_void fw_render(px_surface* psurface, PX_Object* pObject, px_dword elapsed)
{
	struct fw_t* pdesc = PX_ObjectGetDesc(struct fw_t, pObject);
	pdesc->elapsed += elapsed;
	if (pdesc->state == STATE_MOVING)
	{
		for (px_int i = 0; i < 1; i++)
		{
			for (px_int j = 0; j < PX_COUNTOF(pdesc->sparks[i].points); j++)
			{
				px_float delta_x = elapsed * pdesc->sparks[i].points[j].speed_x;
				px_float delta_y = elapsed * pdesc->sparks[i].points[j].speed_y;
				pdesc->sparks[i].points[j].x = pdesc->sparks[i].points[0].x + delta_x - ((px_float)j / PX_COUNTOF(pdesc->sparks[i].points)) * 10 * delta_x;
				pdesc->sparks[i].points[j].y = pdesc->sparks[i].points[0].y + delta_y - ((px_float)j / PX_COUNTOF(pdesc->sparks[i].points)) * 10 * delta_y;

				if (j)
				{
					pdesc->sparks[i].points[j].x += ((px_int)PX_rand() % 127 - 63) * j / 256.0;
					pdesc->sparks[i].points[j].y += ((px_int)PX_rand() % 127 - 63) * j / 256.0;
				}

				pdesc->sparks[i].points[j].size = pdesc->sparks[i].size * (1 - (px_float)j / PX_COUNTOF(pdesc->sparks[i].points));
			}
		}

		if (pdesc->elapsed > 2000)
		{
			for (px_int i = 0; i < PX_COUNTOF(pdesc->sparks); i++)
			{
				for (px_int j = 0; j < PX_COUNTOF(pdesc->sparks[i].points); j++)
				{
					pdesc->sparks[i].points[j].x = pdesc->sparks[i].points[0].x;
					pdesc->sparks[i].points[j].y = pdesc->sparks[i].points[0].y;
				}
			}
			pdesc->state = STATE_TRIGGING;
		}
	}
	else if (pdesc->state == STATE_TRIGGING)
	{
		pdesc->state = STATE_ENERGY_STORING;
	}
	else if (pdesc->state == STATE_ENERGY_STORING)
	{
		pdesc->state = STATE_BLOOMING;

		for (px_int i = 0; i < PX_COUNTOF(pdesc->sparks); i++)
		{
			for (px_int j = 0; j < PX_COUNTOF(pdesc->sparks[i].points); j++)
			{
				pdesc->sparks[i].points[j].x = pdesc->sparks[0].points[0].x;
				pdesc->sparks[i].points[j].y = pdesc->sparks[0].points[0].y;
				pdesc->sparks[i].points[j].speed_x = ((px_int)PX_rand() % 65535 - 32767) / 65536.0 * 1.5;
				pdesc->sparks[i].points[j].speed_y = ((px_int)PX_rand() % 65535 - 32767) / 65536.0 * 1.5;
			}
		}

	}
	else if (pdesc->state == STATE_BLOOMING)
	{
		for (px_int i = 0; i < PX_COUNTOF(pdesc->sparks); i++)
		{
			px_float delta_x = elapsed * pdesc->sparks[i].points[0].speed_x;
			px_float delta_y = elapsed * pdesc->sparks[i].points[0].speed_y;

			pdesc->sparks[i].points[0].speed_x *= 0.995;
			pdesc->sparks[i].points[0].speed_y *= 0.995;

			for (px_int j = 0; j < PX_COUNTOF(pdesc->sparks[i].points); j++)
			{
				pdesc->sparks[i].points[j].x = pdesc->sparks[i].points[0].x + delta_x;
				pdesc->sparks[i].points[j].y = pdesc->sparks[i].points[0].y + delta_y;

				if (j)
				{
					pdesc->sparks[i].points[j].x += ((px_int)PX_rand() % 127 - 63) / 3.0;
					pdesc->sparks[i].points[j].y += ((px_int)PX_rand() % 127 - 63) / 3.0;
					pdesc->sparks[i].points[j].x += ((px_int)PX_rand() % 127 - 63) / 3.0;
					pdesc->sparks[i].points[j].y += ((px_int)PX_rand() % 127 - 63) / 3.0;
				}

				pdesc->sparks[i].points[j].size = pdesc->sparks[i].size / 2; // * (1 - (px_float)j / PX_COUNTOF(pdesc->sparks[i].points));
			}
			if (pdesc->sparks[i].points[0].x < 0 || pdesc->sparks[i].points[0].x > 900)
			{
				pdesc->sparks[i].points[0].speed_x *= -1;
			}
			if (pdesc->sparks[i].points[0].y < 0 || pdesc->sparks[i].points[0].y > 500)
			{
				pdesc->sparks[i].points[0].speed_y *= -1;
			}
			pdesc->sparks[i].points[0].speed_y += 0.02;
		}

		if (pdesc->elapsed > 5000)
		{
			for (px_int i = 0; i < PX_COUNTOF(pdesc->sparks); i++)
			{
				if (PX_rand() % 32 == 0)
				{
					pdesc->sparks[i].size = 0;
				}
			}
		}
	}

	for (px_int i = 0; i < PX_COUNTOF(pdesc->sparks); i++)
	{
		for (px_int j = 0; j < PX_COUNTOF(pdesc->sparks[i].points); j++)
		{
			px_color color = PX_COLOR(255, 255, 255, 255);
			PX_GeoDrawSolidCircle(psurface, (px_int)pdesc->sparks[i].points[j].x, (px_int)pdesc->sparks[i].points[j].y, (px_int)(pdesc->sparks[i].points[j].size), pdesc->sparks[i].color);
		}
	}

	if (pdesc->elapsed > 30000)
	{
		PX_ObjectDelayDelete(pObject);
	}
}

PX_Object *PX_FireworkCreate(px_memorypool *mp, PX_Object *Parent, px_float x,
                             px_float y) {
  // printf("X: %f, Y: %f\n", x, y);
  px_int i;
  struct fw_t desc;
  PX_memset(&desc, 0, sizeof(desc));
  desc.state = STATE_MOVING;
  desc.elapsed = 0;
  desc.start_coord.x = 450 + (px_int)PX_rand() % 255 - 127;
  desc.start_coord.y = 550;
  desc.bloom_coord.x = x;
  desc.bloom_coord.y = 200 - PX_rand() % 64;

  for (px_int i = 0; i < PX_COUNTOF(desc.sparks); i++) {
	desc.sparks[i].color = PX_COLOR(255, PX_rand() % 128 + 64, PX_rand() % 128 + 32, PX_rand() % 128);
  	desc.sparks[i].size = 3;
    for (px_int j = 0; j < PX_COUNTOF(desc.sparks[i].points); j++) {
      desc.sparks[i].points[j].x = desc.start_coord.x;
      desc.sparks[i].points[j].y = desc.start_coord.y;

      desc.sparks[i].points[j].speed_x =
          (desc.bloom_coord.x - desc.start_coord.x) / 2000;
      desc.sparks[i].points[j].speed_y =
          (desc.bloom_coord.y - desc.start_coord.y) / 2000;
    }
  }

  return PX_ObjectCreateEx(mp, Parent, x, y, 0, 0, 0, 0, 0, 0, fw_render, 0,
                           &desc, sizeof(desc));
}
